# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import mock
import contextlib
from nose.tools import eq_
from IPy import IP, IPSet

import prettyip

current_representer = None


def s(*ips):
    "Construct an IPSet from the given strings"
    return IPSet([IP(ip) for ip in ips])


def t(input, output):
    "Test the current representer with the given input and output"
    eq_(sorted(list(current_representer(input))), sorted(output))


@contextlib.contextmanager
def representer(fn):
    "Set the current representer"
    with patched_reresentations_for():
        global current_representer
        current_representer = fn
        try:
            yield
        finally:
            current_representer = None


@contextlib.contextmanager
def patched_reresentations_for():
    """Patch representations_for so that it does not recurse to other
    representers but returns a fixed string"""
    with mock.patch('prettyip.representations_for') as reps_for:
        def fake(ipset, ignore=None):
            ipstrs = ', '.join(str(ip) for ip in ipset.prefixes)
            yield (99.9, 'reps_for({0})'.format(ipstrs))
        reps_for.side_effect = fake
        yield reps_for


def test_singleton():
    with representer(prettyip.singleton):
        yield (t, s('1.2.3.0/24'),
               [(1.0, '1.2.3.0/24')])
        yield (t, s('1.2.3.0/24', '1.1.1.1'),
               [])


def test_empty():
    with representer(prettyip.empty):
        yield (t, s(),
               [(0.0, 'nothing')])
        yield (t, s('1.2.3.0/24', '1.1.1.1'),
               [])


def test_range():
    with representer(prettyip.dashed_range):
        yield (t, s('1.0.0.3', '1.0.0.4/31', '1.0.0.6'),
               [(1.0, '1.0.0.{3-6}')])

        yield (t, s('1.0.0.0/24') - s('1.0.0.0'),
               [(1.0, '1.0.0.{1-255}')])

        # a singleton in the middle
        yield (t, s('1.0.0.0/24') - s('1.0.0.17') - s('1.0.0.19'),
               [(3.25, '1.0.0.{0-16}, 1.0.0.18, 1.0.0.2{0-55}')])

        # a range encompassing all digits, with a very bad score
        yield (t, s('0.0.0.0/5', '8.0.0.0/8', '9.0.0.0/8'),
               [(104.0, '{0.0.0.0-9.255.255.255}')])

        # CIDR range
        yield (t, s('1.0.0.0/24') - s('1.0.0.128'),
               [(1.5, '1.0.0.0/25, 1.0.0.{129-255}')])


def test_except_for():
    with representer(prettyip.except_for):
        yield (t, s('1.0.0.0/20') - s('1.0.1.128'),
               [(100.9, '1.0.0.0/20 except reps_for(1.0.1.128)')])

        yield (t, s('1.0.0.0/17') + s('1.0.128.0/17') - s('1.0.1.128'),
               [(100.9, '1.0.0.0/16 except reps_for(1.0.1.128)')])

        yield (t, s('0.0.0.0/0') - s('172.16.2.0/24'),
               [(100.9, '0.0.0.0/0 except reps_for(172.16.2.0/24)')])


def test_prefix_list():
    with representer(prettyip.prefix_list):
        yield (t, s('1.0.0.0/24', '2.0.0.0/15', '3.0.0.0/8'),
               [(30.0, '1.0.0.0/24, 2.0.0.0/15, 3.0.0.0/8')])


def test_integration():
    def i(input, output):
        eq_(prettyip.pretty_ipset(input), output)
    yield i, s(), 'nothing'
    for bits in range(0, 32):
        yield i, s('0.0.0.0/{0}'.format(bits)), '0.0.0.0/{0}'.format(bits)
    yield i, s('1.0.0.0/8'), '1.0.0.0/8'
    yield i, s('1.2.0.0/16'), '1.2.0.0/16'
    yield i, s('1.2.3.0/24'), '1.2.3.0/24'
    yield i, s('1.2.3.4'), '1.2.3.4'  # no /32 suffix
    yield i, s('1.0.0.0/20') - s('1.0.1.128'), \
        '1.0.0.0/20 except 1.0.1.128'
    yield i, s('1.0.0.0/8') - s('1.0.1.0/24') - s('1.0.9.0/24'), \
        '1.0.0.0/8 except 1.0.1.0/24, 1.0.9.0/24'
    yield i, s('1.0.0.0/8') - s('1.15.0.0/16') + s('1.15.20.0/24'), \
        '1.0.0.0/8 except 1.15.{0.0-19.255}, 1.15.2{1.0-55.255}'


def test_patched():
    prettyip.patch_ipy()
    eq_(str(s('1.0.0.0/25', '1.0.0.128/25')), '1.0.0.0/24')



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from IPy import IP, IPSet
import os

POWERS_OF_2 = dict((2 ** n, n) for n in range(0, 32 + 1))

representers = []


def representer(fn):
    representers.append(fn)
    return fn


@representer
def singleton(ipset):
    if len(ipset.prefixes) == 1:
        yield (1.0, str(ipset.prefixes[0]))


@representer
def empty(ipset):
    if len(ipset.prefixes) == 0:
        yield (0.0, 'nothing')


@representer
def dashed_range(ipset):
    if len(ipset.prefixes) == 0:
        return

    range = None
    ranges = []
    for pfx in ipset.prefixes:
        pfx_int = pfx.int()
        if range is None or pfx_int != range[1] + 1:
            range = [pfx_int, pfx_int + len(pfx) - 1]
            ranges.append(range)
        else:
            range[1] += len(pfx)

    def fmtrange(start_int, end_int):
        # bail out early for singletons..
        if start_int == end_int:
            return 0.25, str(IP(start_int))
        # and for CIDR ranges
        bit_diff_exp = POWERS_OF_2.get((start_int ^ end_int) + 1)
        if bit_diff_exp is not None:
            return 0.5, '{0!s}/{1!s}'.format(IP(start_int), 32 - bit_diff_exp)
        # otherwise make a dashed range
        start, end = str(IP(start_int)), str(IP(end_int))
        prefix = os.path.commonprefix([start, end])
        pfxlen = len(prefix)
        start, end = start[pfxlen:], end[pfxlen:]
        # calculate th score for this range, penalizing for shorter
        # prefixes (longer dashed ranges) and for not ending on a dot
        score = 4.0 - prefix.count('.')
        if not prefix:
            # wrapping {..} around the whole ipset is not readable!
            score += 100.0
        elif prefix[-1] != '.':
            score += 1.0
        return score, '{0}{{{1}-{2}}}'.format(prefix, start, end)

    scored_formats = [fmtrange(s, e) for s, e in ranges]
    score = sum(range_score for (range_score, _) in scored_formats)
    yield score, ", ".join(range_fmt for (_, range_fmt) in scored_formats)


@representer
def except_for(ipset):
    # this isn't worth it for simple sets
    if len(ipset.prefixes) < 3:
        return

    def mkip(net, prefixlen):
        ip = IP(net)
        ip._prefixlen = prefixlen
        return ip

    def ipset_in_ip(ipset, ip):
        for prefix in ipset.prefixes:
            if prefix not in ip:
                return False
        return True

    lower = 0
    smallest_containing = IP('0.0.0.0/0')
    for bitlength in range(1, 24):
        low_half = mkip(lower, bitlength)
        # this doesn't work :(
        if ipset_in_ip(ipset, low_half):
            smallest_containing = low_half
            continue
        lower += 1 << (32 - bitlength)
        high_half = mkip(lower, bitlength)
        if ipset_in_ip(ipset, high_half):
            smallest_containing = high_half
            continue
        break

    # if we found a CIDR block containing the whole IPSet,
    # try inverting it
    if smallest_containing:
        prefix = "%s except " % (smallest_containing)
        inverted = IPSet([smallest_containing]) - ipset
        for score, rep in representations_for(inverted, ignore=[except_for]):
            yield score + 1, prefix + rep


@representer
def prefix_list(ipset):
    yield (10.0 * len(ipset.prefixes),
           ", ".join(str(ip) for ip in ipset.prefixes))


def representations_for(ipset, ignore=None):
    for representer in representers:
        if ignore and representer in ignore:
            continue
        for score, rep in representer(ipset):
            yield score, rep


def pretty_ipset(ipset):
    best = None
    best_score = None
    for score, rep in representations_for(ipset):
        if best_score is None or score < best_score:
            best = rep
            best_score = score

    return best


def patch_ipy():
    IPSet.__str__ = pretty_ipset



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# based on https://github.com/pypa/sampleproject

from setuptools import setup

import os.path

setup(
    name='prettyip',
    version='1.0.4',
    description="Pretty-print IPy's IPSets ",
    long_description=open(os.path.join(os.path.dirname(__file__), 'README.rst')).read(),
    url='https://github.com/djmitche/prettyip',
    author='Dustin J. Mitchell',
    author_email='dustin@mozilla.com',
    license='MPLv2',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: System :: Networking',
        'Topic :: Internet',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
    ],
    keywords='ipy ip pretty-print',
    py_modules=['prettyip', 'test_prettyip'],
    install_requires=['IPy'],
    extras_require={
        'test': ['nose', 'mock'],
    },
)
